/*

File: mrw.gp
Author: Greg Zaverucha
Date: 2007-04-17


This file contains two user-functions

1. mrw(f)	computes the modular reduction weight of a polynomial f

2. exhaustive_polsearch(d,A,maxmrw=0) 
exhaustively searches all polynomials with coefficients 
of absolute value less than A and degree d
warning!!!
uses: d*(2*A+1) space
and time: O( (2*A + 1)^d )

*/


mrw(f)=
{
    local(d, X,i,j,Z_max,Y_max,Z_j,Y_j);

    d = poldegree(f);
    X = compute_X(f);
    \\print("x = ");
    \\print(X);

    Z_max = Y_max = 0;
    for(j = 1, d,
	Y_j = Z_j =0;
	for(i = 1, d,
	    if(X[i,j] > 0,  Y_j += X[i,j];);
	    if(X[i,j] < 0,  Z_j += abs(X[i,j]););
	);
    
	if(Y_j  > Y_max, Y_max = Y_j;);
	if(Z_j  > Z_max, Z_max = Z_j;);
    );

    return(Z_max + Y_max);
}

compute_X(f)=
{
    local(d,c,X,i,j,dm1);

    d = poldegree(f);
    dm1 = d-1;
    X = matrix(d,d);
    c = vector(d);

    /* fill the coeff vector */
    j = 1;
    forstep(i = poldegree(f) - 1, 0, -1,
	c[j] = -polcoeff(f,i);
	j++;
    );
    \\print("coeffs = ",c);

    /* fill X */
    for(j = 1, d,  /* first row */
	X[1,j] = c[d-j + 1];
    );

    for(i = 2, d,
	X[i,1] = X[i-1,d]*c[d];
	for(j = 2, d,
	    X[i,j] = X[i-1,j-1] + X[i-1,d]*c[d-(j-1)];
	);
    );

    return(X);
}

{
print_X(f)=
    local(d,c,X,i,j,dm1);

    d = poldegree(f);
    dm1 = d-1;
    X = matrix(d,d);
    c = vector(d);

    /* fill the coeff vector */
    j = 1;
    forstep(i = poldegree(f) - 1, 0, -1,
	c[j] = -polcoeff(f,i);
	j++;
    );
    \\print("coeffs = ",c);

    /* fill X */
    for(j = 1, d,  /* first row */
	X[1,j] = c[d-j + 1];
	print1(c[d-j + 1], " " );
    );
    print1("\n");

    for(i = 2, d,
	X[i,1] = X[i-1,d]*c[d];
	for(j = 2, d,
	    X[i,j] = X[i-1,j-1] + X[i-1,d]*c[d-(j-1)];
	    print1(X[i-1,j-1] + X[i-1,d]*c[d-(j-1)] , " ");
	);
	print1("\n");
    );

    return(X);
}


/*
    see docs at top of file
*/
exhaustive_polsearch(d,A,maxmrw=0)=
{
    local(c,C,v,f,i);
    c = vector(2*A+1, X, -A+X-1); \\ vector of possible coefficients
    C = c;
    for(i = 2, d, 
	C = concat(C, c); \\ we'll loop over all d-subsets
    );

    forvec( v=vector(d, i,[1,length(C)]),   \\ loop snippet from the pari-gp faq
	f = vec2poly( vecextract(C,v) );
	m = mrw(f);
	if( maxmrw && m < maxmrw,
	    print(m, "  ", f);
	);
	if(!maxmrw,
	    print(m, "  ", f);
	);
    );

}


/*
    turns the vector v into a polynomial
    f = v[d]x^d + .... + v[1]
*/
vec2poly(v)=
{
    local(d,f);

    d = length(v);
    f = t^d;
    for(i = 1, d-1,
	f +=  v[i]*t^(d-i) ;
    );
    f += v[d];

    return(f);
}

